# Copyright (c) 2007 Google Inc.
require 'sketchup.rb'
require'CityGML/AttributeRescue.rb'

module FHGelsenkirchen

  class DataInitializer

    def initialize(use_def_if_available)
      #Sketchup.active_model.start_operation("Attribute Edits");
      ss = Sketchup.active_model.selection
      if(ss.empty? || ss.count != 1)
        $entity = Sketchup.active_model;
      else
        $entity = ss[0];
        if($entity.class == Sketchup::ComponentInstance)
          if(use_def_if_available)
            $entity = $entity.definition
          elsif(IDYES == UI::messagebox("Use Definition?", MB_YESNO))
            $entity = $entity.definition
          end
        end
      end
      initStandardTabs();
    end

    def initTextAttrib(dict,name)
      $entity.set_attribute(dict,name,dict+":"+name) if(nil == $entity.get_attribute(dict,name))
    end
    
    def initLinkAttrib(dict,name,url)
      $entity.set_attribute(dict,name,url) if(nil == $entity.get_attribute(dict,name))
    end
    
    def initAuthorTab()
      initTextAttrib("AuthorInfo","name");
      initLinkAttrib("AuthorInfo","email", "mailto://someone@somewhere.com");
      initLinkAttrib("AuthorInfo","link", "http://sketchup.google.com");
    end
    
    def initProductTab()
      initTextAttrib("ProductInfo","brand");
      initTextAttrib("ProductInfo","brand_link");
      initTextAttrib("ProductInfo","link");
      initTextAttrib("ProductInfo","image_link");
      initTextAttrib("ProductInfo","id");
      initTextAttrib("ProductInfo","product_type");
      initTextAttrib("ProductInfo","specification_code");
      initTextAttrib("ProductInfo","specification_standard");
    end
    def initModelTab()
      initTextAttrib("ModelInfo","title");
      initTextAttrib("ModelInfo","description");
      initTextAttrib("ModelInfo","tags");
      initLinkAttrib("ModelInfo","link","http://sketchup.google.com/3dwarehouse");
      initTextAttrib("ModelInfo","warehouse_id");
      initTextAttrib("ModelInfo","last_modified");
    end

	
    
    def initStandardTabs()
      initAuthorTab();
      initModelTab();
      initProductTab();
    end
  end

  class AttribManagerBase

    def initialize(use_def_if_available)
      @modified = false;
      Sketchup.active_model.start_operation("Attribute Edits");
      ss = Sketchup.active_model.selection
      if(ss.empty? || ss.count != 1)
        $entity = Sketchup.active_model;
      else
        $entity = ss[0];
        if($entity.class == Sketchup::ComponentInstance)
          if(use_def_if_available)
            $entity = $entity.definition
          elsif(IDYES == UI::messagebox("Use Definition?", MB_YESNO))
            $entity = $entity.definition
          end
        end
      end
      cleanser = AttributeCleanser.new($entity);
      cleanser.clean();
        

      closed = false;
       
      $dlg = UI::WebDialog.new("Attribute edit", true, "AM");
      $dlg.set_file(File.dirname(__FILE__) + "/AttribEditor.html");
      $dlg.add_action_callback("on_cancel") {|d,p| closed=true;Sketchup.active_model.abort_operation(); $dlg.close() }
      $dlg.add_action_callback("on_accept") {|d,p| closed=true;Sketchup.active_model.commit_operation(); $dlg.close() }
      $dlg.add_action_callback("on_close") {|d,p| $dlg.close() }
      $dlg.add_action_callback("on_delete_attribute") {|d,p| deleteAttribute(d, p); }
      $dlg.add_action_callback("on_edit_attribute") {|d,p| editAttribute(d, p); }
      $dlg.add_action_callback("on_add_attribute") {|d,p| addAttribute(d, p); }
      $dlg.add_action_callback("on_delete_dictionary") {|d,p| deleteDictionary(d, p); }
      $dlg.add_action_callback("on_add_dictionary") {|d,p| addDictionary(d, p); }
      $dlg.add_action_callback("create_standard_categories") {|d,p| createStandardCategories(d, p); }
      $dlg.add_action_callback("createlist") {|d,p|
     
      }
        

      $dlg.set_size 600, 400;
      $dlg.set_on_close {
        check_save() if(!closed)
      }
      if($entity.class == Sketchup::Face)
        showDialog("face")
      elsif($entity.class == Sketchup::ComponentDefinition or $entity.class == Sketchup::Group)
        showDialog("group")
      end
    end

    def check_save()
      if(@modified)
        save = UI::messagebox("Save changes?", MB_YESNO);
        if(save==IDNO)
          Sketchup.active_model.abort_operation();
        else
          Sketchup.active_model.commit_operation();
        end
      else
        Sketchup.active_model.abort_operation();
      end
    end
    
    def createStandardCategories(d,p) 
      @modified = true;
      DataInitializer.new($entity);
      json = getJSONDictionaries();
      $dlg.execute_script("initJson('" + json + "', true)");
    end
    
    def fromJSON(str)
      j = Hash.new()
      status = 0
      name = ""
      value = ""
      begin
        str.each_byte do |c|
          case status
          when 0
            status = 1 if(c.chr == "\"")
          when 1
            if(c.chr != "\"")
              name << c
            else
              status = 2
            end
          when 2
            status = 3 if(c.chr == "\"")
          when 3
            if(c.chr != "\"")
              value << c
            else
              if(name == "attrib_name" || name == "attrib_value")
                value = FHGelsenkirchen::str_convert(value)
              end
              j[name] = value
              name = ""
              value = ""
              status = 0
            end
          else
          end
        end

        #new_str = str.gsub(/(["'])\s*:\s*(['"0-9tfn\[{])/){"#{$1}=>#{$2}"}
       #new_str = str.gsub(/[:]/,'=>')
       #puts "#{new_str}\n"
       #ret_val = eval(new_str)
      rescue Exception => exc
       puts exc
      end
      return j
    end
    
	
    def deleteAttribute(d, p)
      if(p == nil)
        return
      end
      p.chop!
      @modified = true;
      # extract the dict and name from the parameters
      dict_and_name = fromJSON(p);
      dict = $entity.attribute_dictionaries[dict_and_name["dict_name"]];
      if(dict != nil)
        dict.delete_key dict_and_name["attrib_name"];
      end
    end
    
    def editAttribute(d, p)
      if(p == nil)
        return
      end
      p.chop!
      @modified = true;
      params = fromJSON(p);
      dict = $entity.attribute_dictionaries[params["dict_name"]];
      #params["old_attrib_name"] = params["dict_name"]
      if(dict != nil)
        # delete the existing attribute
        dict.delete_key params["attrib_name"];
        # add it with the new name/value
        dict[ params["attrib_name"] ] = params["attrib_value"];
      end
    end

    /def check_dic_name(params)
      entity = Sketchup.active_model.selection[0]
      if(entity.class == Sketchup::Face)
        case params["attrib_name"]
        when "PolygonID"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "LinearRingID"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        else
          dict = $entity.attribute_dictionaries["Generic attribute"]
        end
      else
        case params["attrib_name"]
        when "id"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "class"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "function"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "usage"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "yearOfConstruction"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "roofType"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "measuredHeight"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "storeysAboveGround"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "storeysBelowGround"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "storeyHeightsAboveGround"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "storeyHeightsBelowGround"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "creationDate"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        when "terminationDate"
          dict = $entity.attribute_dictionaries["Standard attribute"]
        else
          dict = $entity.attribute_dictionaries["Generic attribute"]
        end
      end
      return dict
    end/

    def addAttribute(d, p)
      if(p == nil)
        return
      end
      p.chop!
      @modified = true;
      params = fromJSON(p);
       
      dict = $entity.attribute_dictionaries[params["dict_name"]]   
      if(dict != nil)
        dict[ params["attrib_name"] ] = params["attrib_value"]; 
      end
    end
    
    def deleteDictionary(d, p)
      @modified = true;
      $entity.attribute_dictionaries.delete p;
    end

    def addDictionary(d, p)
      @modified = true;
      $entity.attribute_dictionary p, true;
    end

    # escapges the characters in the given string so that the
    # resulting string can be used as an argument to a javascript
    # function.
    def escapeChars(s)
      s.gsub('"','\\\\\\"').gsub("'","\\\\'");
    end

    def getJSONDictionaries()
      json = '{"dictionaries":[';
      dicts = $entity.attribute_dictionaries
      if(dicts != nil)
        dicts.each do |dict|
          json += '{"name":"' + dict.name + '","attributes":[';                
          dict.each_key do |attribName|
            json += '{';
            json += '"name":"' + FHGelsenkirchen::str_convert_back(attribName.to_s) + '",';
            if(dict[attribName].class != Geom::Transformation)
              json += '"value":"' + FHGelsenkirchen::str_convert_back(escapeChars(dict[attribName].to_s)) + '"},';
            else
              json += '"value":"' + FHGelsenkirchen::str_convert_back(escapeChars(dict[attribName].to_a.to_s)) + '"},';
            end
          end
          json += ']},'
        end
      end
      json += "]}";
      json.gsub!(/,/,"#COMMA#");
      return json;
    end
    
    def showDialog(type)
      $dlg.show_modal() {
        script = "document.getElementById(\"theTypeField\").value=\"#{type}\";"
        $dlg.execute_script(script);
        json = getJSONDictionaries();
        $dlg.execute_script("initJson('" + json + "')");
      }
    end
    
  end

  class AttribManagerAll < AttribManagerBase
  	def initialize()
      super(false)
    end
  end

  class AttribManagerProduct < AttribManagerBase
  	def initialize()
      super(true)
    end
  end

  class AttributeCleanser
    def initialize(ent)
      $ent = ent
    end
    
    def clean()
      dicts = []
      toDelete = []
      dictNames = []
      ads = $ent.attribute_dictionaries;
      return if ads == nil;
      i = 0
      ads.each do |d|
        if(dictNames.index(d.name) == nil)
          dicts.push(d)
          dictNames.push(d.name)
        else
          toDelete.push(d)
        end
        i = i + 1
      end
        
      toDelete.each do |d|
        ads.delete d
      end
    end
  end

  $helpHTML=<<EOHELP
    <html>
    <body>
    <b>Help for Attribute Manager functions.</b>
    <p>
    This plugin helps you manage the attributes found on the model and 
    entities contained within the model.  
    <p>
    It is accessed in two ways: 
    <ol>
      <li>Via the main menus: <span style="font-family:courier">Tools &gt; Attributes</span>
      The options here apply to the entire model.
      <li>Via context (right click) menus: <span style="font-family:courier">Attributes</span>
      The options here apply to the <em>single</em> selected item.
    </ol>
    <p>
    There are 3 operations possible from the menus - 
    show the attributes, cleanse the attributes of duplicate categories, and 
    make the standard categories.
    <p>
    The <b>Show</b> option will launch a modal dialog that allows you to create and delete 
    attribute categories and add/edit/delete individual attributes.
    <p>
    If an attribute value starts with "http://" or "mailto://" then the value will be displayed 
    without the prefix, but with the link automatically created.
    </body>
    </html>
EOHELP

  subMenu = UI::menu("Tools").add_submenu("Attribute");
  subMenu.add_item("Edit") {
    AttribManagerAll.new
  }

  #subMenu.add_item("Make standard categories") {
  #	DataInitializer.new(true)
  #	}
  #subMenu.add_item("Lösche Duplikate") {
  #    # make sure there is only one of each dictionary of a given name
  #    ac = AttributeCleanser.new(Sketchup.active_model);
  #    ac.clean();
  #}
  def setgmlstandardattribute(dictionary, name)
    dictionary[name] = "" if (!dictionary[name])
  end

  #subMenu.add_separator
  #subMenu.add_item("Hilfe") {
  #	dlg = UI::WebDialog.new("Attribute Manager Hilfe", true, "AMHelp")
  #    dlg.set_html($helpHTML)
  #    dlg.show
  #	}



  UI::add_context_menu_handler do |menu|
    $wallsurface = false
    $roofsurface = false
    $groundsurface = false
    $relieffeature = false
    if(Sketchup.active_model.selection.count >= 1)
      menu.add_separator
      entity = Sketchup.active_model.selection[0]
      if(Sketchup.active_model.selection.count == 1)
        if(entity.class == Sketchup::Group)
          @selected_type = "Group"
          attrdicts = entity.attribute_dictionaries
          if(attrdicts != nil)
            names = []
            attrdicts.each do |dict|
              names << dict.name if(dict.name != "Standard attribute" and dict.name != "Generic attribute")
            end
            names.each { |name|
              attrdicts.delete(name)
            }
          end
          dictionary = entity.attribute_dictionary("Standard attribute", true)
          entity.attribute_dictionary("Generic attribute", true)
        elsif(entity.class == Sketchup::ComponentInstance)
          @selected_type = "Group"
          definition = entity.definition
          attrdicts = definition.attribute_dictionaries
          if(attrdicts != nil)
            names = []
            attrdicts.each do |dict|
              names << dict.name if(dict.name != "Standard attribute" and dict.name != "Generic attribute")
            end
            names.each { |name|
              attrdicts.delete(name)
            }
          end
          definition.attribute_dictionary("Standard attribute", true)
          definition.attribute_dictionary("Generic attribute", true)
        elsif(entity.class == Sketchup::Face)
          @selected_type = "Face"
          dictionary = entity.attribute_dictionary("Standard attribute", true)
         # entity.attribute_dictionary("Generic attribute", true)    # Auskommentiert: Kein Generic Tab für Faces
        end
        subMenu = menu.add_submenu("Attribute");

        subMenu.add_item("Edit") {
          AttribManagerProduct.new;
        }
        subMenu.add_item("Copy") {
          AttributeRescue.new.copy_attribute(Sketchup.active_model.selection.first)
        }
        subMenu.add_item("Insert") {
          AttributeRescue.new.paste_attribute(Sketchup.active_model.selection.first)
        }
      end
      if(Sketchup.active_model.selection.count >= 1)
        if(FHGelsenkirchen::hasface?(Sketchup.active_model.selection))
          subMenu = menu.add_submenu("CityGML Surfacetype");

          itemground = subMenu.add_item("GroundSurface") {
            FHGelsenkirchen::setsurfacetype(Sketchup.active_model.selection, "GroundSurface")
          }
          itemroof = subMenu.add_item("RoofSurface") {
            FHGelsenkirchen::setsurfacetype(Sketchup.active_model.selection, "RoofSurface")
          }
          itemwall = subMenu.add_item("WallSurface") {
            FHGelsenkirchen::setsurfacetype(Sketchup.active_model.selection, "WallSurface")
          }
         # itemrelief = subMenu.add_item("ReliefFeature"){
         #   FHGelsenkirchen::setsurfacetype(Sketchup.active_model.selection, "ReliefFeature")
        #  }
          itemnone = subMenu.add_item("None") {
            FHGelsenkirchen::setsurfacetype(Sketchup.active_model.selection, nil)
          }
          surfaceid = subMenu.add_item("BoundarySurfaceType ID") {
            idhash = FHGelsenkirchen::setsurfaceid(Sketchup.active_model.selection, Hash.new())
            defaults = []
            prompts = ["SurfaceType ID"]
            if(idhash.size == 1)
              idhash.each_pair {|key, value|
                if(value.size == 1)
                  value.each_pair {|key2, value2|
                    defaults = [key2]
                  }
                end
              }

              input = UI.inputbox prompts, defaults, "BoundarySurfaceType ID"
              FHGelsenkirchen::setid(Sketchup.active_model.selection, input[0]) if(input)
            else
              UI.messagebox("The selected faces are not on the same layer")
            end
          }
          subMenu.set_validation_proc(itemground) {

            if $groundsurface == true
              MF_CHECKED
            else $groundsurface == false
              MF_UNCHECKED
            end
          }
          subMenu.set_validation_proc(itemroof) {

            if $roofsurface == true
              MF_CHECKED
            else $roofsurface == false
              MF_UNCHECKED
            end
          }
          subMenu.set_validation_proc(itemwall) {
            if $wallsurface == true
              MF_CHECKED
            else $wallsurface == false
              MF_UNCHECKED
            end
          }
          #subMenu.set_validation_proc(itemrelief) {
          #  if $relieffeature == true
           #   MF_CHECKED
           # else $relieffeature == false
           #   MF_UNCHECKED
           # end
         # }
        end
        entities = Sketchup.active_model.selection
        typehash = Hash.new()
        FHGelsenkirchen::findtype(entities, typehash)

        if(typehash.size == 1)
          if(typehash["WallSurface"] != nil)
            $wallsurface = true
            $groundsurface = false
            $roofsurface = false
            $relieffeature = false
          elsif(typehash["RoofSurface"] != nil)
            $wallsurface = false
            $groundsurface = false
            $roofsurface = true
            $relieffeature = false
          elsif(typehash["GroundSurface"] != nil)
            $wallsurface = false
            $groundsurface = true
            $roofsurface = false
            $relieffeature = false
          elsif(typehash["ReliefFeature"] != nil)
            $wallsurface = false
            $groundsurface = false
            $roofsurface = false
            $relieffeature = true
          end
        end

        #Standard attribute

      end








      #subMenu.add_item("Erzeuge standard Kategorien") {
      #    DataInitializer.new(true)
      #}

      #subMenu.add_item("Lösche Duplikate") {
      #    AttributeCleanser.new(Sketchup.active_model.selection.first);
      #}
    end
  end
  def FHGelsenkirchen::setid(entities, id)
    entities.each do |e|
      if(e.class == Sketchup::Face)
        begin
          #e.set_attribute "Standard attribute", "SurfaceType", type
          e.set_attribute("Standard attribute", "BoundarySurfaceType ID", id)
        rescue => err
          puts err
        end
      elsif(e.class == Sketchup::Group)
        setid(e.entities, id)
      elsif(e.class == Sketchup::ComponentInstance)
        setid(e.definition.entities, id)
      end
    end
  end

  def FHGelsenkirchen::setsurfacetype(entities, type)
    $wallsurface = false
    $roofsurface = false
    $groundsurface = false
    $relieffeature = false
    entities.each do |e|
      if(e.class == Sketchup::Face)
        if(type != nil)
          begin
            #e.set_attribute "Standard attribute", "SurfaceType", type
            e.layer= type
          rescue => err
            e.layer= Sketchup.active_model.layers.add(type)
          end
        else
          begin
            #e.delete_attribute "Standard attribute", "SurfaceType"
            face.layer= "No Surfacetype"
          rescue => err
            e.layer= Sketchup.active_model.layers.add("No Surfacetype")
          end
        end
      elsif(e.class == Sketchup::Group)
        setsurfacetype(e.entities, type)
      elsif(e.class == Sketchup::ComponentInstance)
        setsurfacetype(e.definition.entities, type)
      end
    end
  end

  def FHGelsenkirchen::setsurfaceid(entities, hash)
    entities.each do |e|
      if(e.class == Sketchup::Face)
        id = e.get_attribute("Standard attribute", "BoundarySurfaceType ID", "")
        if(hash[e.layer.name] == nil)
          hash[e.layer.name] = Hash.new
        end
        if(hash[e.layer.name][id] == nil)
          hash[e.layer.name][id] = 1
        else
          hash[e.layer.name][id] += 1
        end
      elsif(e.class == Sketchup::Group)
        setsurfaceid(e.entities, hash)
      elsif(e.class == Sketchup::ComponentInstance)
        setsurfaceid(e.definition.entities, hash)
      end
    end
    return hash
  end

  def FHGelsenkirchen::hasface?(entities)
    entities.each do |e|
      if(e.class == Sketchup::Face)
        return true
      elsif(e.class == Sketchup::Group)
        return hasface?(e.entities)
      elsif(e.class == Sketchup::ComponentInstance)
        return hasface?(e.definition.entities)
      end
    end
    return false
  end
  
  def FHGelsenkirchen::findtype(entities, typehash)
    entities.each do |e|
      if(e.class == Sketchup::Face)
        value = nil #= e.get_attribute "Standard attribute", "SurfaceType", nil
        #value = e.layer.name
        #if(e.layer.name != value)
        if(e.layer.name == "GroundSurface" or e.layer.name == "WallSurface" or e.layer.name == "RoofSurface" or e.layer.name == "ReliefFeature" or e.layer.name == "No Surfacetype")
          #e.set_attribute "Standard attribute", "SurfaceType", e.layer.name
          value = e.layer.name
        end
        #end
        if(value != nil)
          if(typehash[value] == nil)
            typehash[value] = 1
          else
            typehash[value] += 1
          end
        end
      elsif(e.class == Sketchup::Group)
        findtype(e.entities, typehash)
      elsif(e.class == Sketchup::ComponentInstance)
        findtype(e.definition.entities, typehash)
      end
    end
  end

  def FHGelsenkirchen::str_convert(s)
    if(s == nil)
      return nil
    end
    str = s.gsub(/\xD6/,$Oe.to_s)
    str = str.gsub(/\xF6/,$oe.to_s)
    str = str.gsub(/\xDC/,$Ue.to_s)
    str = str.gsub(/\xFC/,$ue.to_s)
    str = str.gsub(/\xC4/,$Ae.to_s)
    str = str.gsub(/\xE4/,$ae.to_s)
    str = str.gsub(/\xDF/,$sz.to_s)
    str = str.gsub(/\x80/,$euro.to_s)
    return str
  end

  def FHGelsenkirchen::str_convert_back(s)
    if(s == nil)
      return nil
    end
    str = s.gsub($Oe.to_s,'Ö')
    str = str.gsub($oe.to_s,'ö')
    str = str.gsub($Ue.to_s,'Ü')
    str = str.gsub($ue.to_s,'ü')
    str = str.gsub($Ae.to_s,'Ä')
    str = str.gsub($ae.to_s,'ä')
    str = str.gsub($sz.to_s,'ß')
    str = str.gsub($euro.to_s,'€')
    return str
  end

  def FHGelsenkirchen::str_filter(s)
    if(s == nil)
      return nil
    end
    str = s.gsub(/\xD6/,'Ö')
    str = str.gsub(/\xF6/,'ö')
    str = str.gsub(/\xDC/,'Ü')
    str = str.gsub(/\xFC/,'ü')
    str = str.gsub(/\xC4/,'Ä')
    str = str.gsub(/\xE4/,'ä')
    str = str.gsub(/\xDF/,'ß')
    str = str.gsub(/\x80/,'€')
    return str
  end
end