require 'CityGML/GMLParser'
require 'CityGML/GMLOptionsDialog'
require 'CityGML/bigdecimal'

module FHGelsenkirchen
  #Hauptklasse zum einladen von CityGML
  class Citygml

    #Eigentlich überflüssig, aber räumt alle Arrays auf
    def clearmemory()
      @entities = nil
      @minX = nil
      @minY = nil
      @minZ = nil
      @model = nil
      @pts = nil
      @selection = nil
      @tmpent = nil
      @tmpgroup = nil
      @idhash = nil
      @ptshash = nil
      @surfacetypes = nil
      puts "Number of surfaces: #{@gesamtanzahl}"
      puts "Non-Planar surfaces: #{@nichtplanar}"
      GC.start()
    end

    #Setzt Textur auf ein Face.
    #Parameter 1: Face
    #Parameter 2: Point-Array, in dem alle Koordinaten gespeichert sind
    #Parameter 3: Texturkoordinaten-Array
    def setTexture face, pts, texturcords
      begin
        #Doppelte Koordinaten löschen
        pts.uniq!
        texturcords.uniq!
        #Wenn Anzahl Texturkoordinaten und Koordinaten ungleich ist, ist CityGML-Datei fehlerhaft.
        #Dann wird keine Textur angezeigt
        if(pts.length  != texturcords.length)
          puts "Number of texture coordinates #{texturcords.length} is not even Number of vertices #{pts.length}"
          begin
            puts "PolygonID: #{face.get_attribute("Standard attribute", "PolygonID")}"
          rescue =>e
          end
          puts "Polygon is displayed without textures."
          color = face.material.texture.average_color
          face.material.texture = nil
          face.material.color = color
          return
        end
        cords = []
        texturcords.each do |v|
          cords << [v[0], v[1]]
        end
        if(@switch)
          cords.reverse!
        end
        ptsarray = []
        #Funktion zum Setzen von Texturkoordinaten kann nur 4 Punkte als Übergabeparameter verarbeiten
        #Gibt es mehr als 4 Punkte, müssen welche ausgewählt werden.
        #Am Besten die vier am weitesten voneinander liegenden Punkten
        if(pts.length() > 4)
          #Setzt Werte für Rechts Oben/Unten und Links Oben/Unten
          #Es wird jeweils der größte und kleinste Wert für X und Y gesucht
          
          for i in 0..(cords.length() - 1)
            ro_x = cords[i][0] if(ro_x == nil or cords[i][0] > ro_x)
            ro_y = cords[i][1] if(ro_y == nil or cords[i][1] > ro_y)
            lu_x = cords[i][0] if(lu_x == nil or cords[i][0] < lu_x)
            lu_y = cords[i][1] if(lu_y == nil or cords[i][1] < lu_y)
          end
          #Mit den jeweils 2 Koordinaten für X und Y werden jetzt weitere Punkte
          #berechnet, so wird nen Rechteck erzeugt
          ru_x = ro_x
          ru_y = lu_y
          lo_x = lu_x
          lo_y = ro_y
          abs_ro = []
          abs_ru = []
          abs_lu = []
          abs_lo = []
          for i in 0..(cords.length() - 1)
            abs_ro << [(ro_x - cords[i][0]) * (ro_x - cords[i][0]) + (ro_y - cords[i][1]) * (ro_y - cords[i][1]), i]
            abs_ru << [(ru_x - cords[i][0]) * (ru_x - cords[i][0]) + (ru_y - cords[i][1]) * (ru_y - cords[i][1]), i]
            abs_lu << [(cords[i][0] - lu_x) * (cords[i][0] - lu_x) + (cords[i][1] - lu_y) * (cords[i][1] - lu_y), i]
            abs_lo << [(cords[i][0] - lo_x) * (cords[i][0] - lo_x) + (cords[i][1] - lo_y) * (cords[i][1] - lo_y), i]
          end
          n = abs_ro.length() -1
          abs_ro.sort!
          abs_lo.sort!
          abs_ru.sort!
          abs_lu.sort!
          pos_ru = abs_ru[0][1]
          pos_lu = abs_lu[0][1]
          pos_ro = abs_ro[0][1]
          pos_lo = abs_lo[0][1]
          #Überprüfung, dass ein Eckpunkt nicht doppelt gewählt wird
          if(pos_lu != pos_ru and pos_lu != pos_ro and pos_lu != pos_lo)
            ptsarray << Geom::Point3d.new(pts[pos_lu][0], pts[pos_lu][1], pts[pos_lu][2])
            ptsarray << Geom::Point3d.new(cords[pos_lu][0], cords[pos_lu][1])
          end

          if(pos_ru != pos_ro and pos_ru != pos_lo)
            ptsarray << Geom::Point3d.new(pts[pos_ru][0], pts[pos_ru][1], pts[pos_ru][2])
            ptsarray << Geom::Point3d.new(cords[pos_ru][0], cords[pos_ru][1])
          end

          if(pos_ro != pos_lo)
            ptsarray << Geom::Point3d.new(pts[pos_ro][0], pts[pos_ro][1], pts[pos_ro][2])
            ptsarray << Geom::Point3d.new(cords[pos_ro][0], cords[pos_ro][1])
          end

          ptsarray << Geom::Point3d.new(pts[pos_lo][0], pts[pos_lo][1], pts[pos_lo][2])
          ptsarray << Geom::Point3d.new(cords[pos_lo][0], cords[pos_lo][1])
        else
          for i in 0..(pts.length() - 1)
            ptsarray << Geom::Point3d.new(pts[i][0] , pts[i][1], pts[i][2])
            ptsarray << Geom::Point3d.new(cords[i][0], cords[i][1])
          end
        end

        begin
          #raise "textur nicht vorhanden" if(Sketchup.active_model.materials[texturename] == nil)
          #face.material = texturename
          #puts ptsarray
          face.position_material(face.material, ptsarray, true)
        rescue => e
          puts e
          #Wenn hier rein, dann wurden Koordinaten falsch gewählt
        end
      rescue => e
        puts e
        puts e.backtrace
      end
    end


    #Bekommt appearanceMember übergeben
    #Iterriert über alle Elemente und setzt Texturen und Material
    def setMaterials(appearanceMember)
      #UI.messagebox "Material start", MB_OK
      matnr = 1
      return if(appearanceMember == nil)
      appearanceMember.SurfaceDataMembers.each do |surfaceDataMember|
        material = surfaceDataMember.Material
        texture = surfaceDataMember.Texture
        if(material)  #Es gab ein Material
          matnr += 1
          #Im Target stehen IDs der LinearRings
          material.Targets.each do |target|
            begin
              target.strip! #Entfernt Whitespaces, die ab und zu beim Parsen vorkommen
              faces = @idhash.fetch(target)
              faces.each do |face|
                begin
                  #Überprüfen, ob Material schon angelegt ist, sonst wird es angelegt
                  #Bei Materialien wird der Materialname nach den Farbwerten aus der CityGML Datei benannt
                  #Beispiel: Farbwerte Rot=1, Grün=0.5, Blau=0.25 erzeugt Texturnamen "1 0.5 0.25"
                  #Dadurch erhalten alle Flächen mit dem selben Material auch das selbe Materialobjekt
                  if(Sketchup.active_model.materials["#{material.DiffuseColor[0]} #{material.DiffuseColor[1]} #{material.DiffuseColor[2]}"] == nil)
                    face.material = "#{material.DiffuseColor[0]} #{material.DiffuseColor[1]} #{material.DiffuseColor[2]}"
                  else
                    m = @skpmaterials.add("#{material.DiffuseColor[0]} #{material.DiffuseColor[1]} #{material.DiffuseColor[2]}")
                    color = Sketchup::Color.new()
                    #CityGML Farben sind von 0 - 1, SketchUp zwischen 0 - 255
                    color.red= material.DiffuseColor[0].to_f * 255
                    color.green= material.DiffuseColor[1].to_f * 255
                    color.blue= material.DiffuseColor[2].to_f * 255
                    color.alpha= material.Transparency.to_f * 255
                    m.color= color
                    face.material = m
                  end
                rescue => e

                end
              end
            rescue => e
              puts e
            end
          end
        elsif(texture) # Es gab eine Textur
          #TargetTextureCoordinates speichert zu jeder ID die Texturkoordinaten
          texture.TargetTextureCoordinates.each_pair do |key, value|
            begin
              faces = @idhash.fetch(key)  #Faces aus der HashMap holen
              pts = @ptshash.fetch(key) #Für Texturkoordinaten werden auch Koordinaten benötigt
              begin
                cleanpts = @cleaned_pts.fetch(key)
                cleanpts.each do |pt|
                  value.delete_at(pt)
                end
              rescue => e
              end
              faces.each do |face|
                setTexture(face, pts, value, texture.ImageURI)
              end
            rescue => e
              #puts e
            end
          end
        end
      end
    end

    #Setzt Attribute auf ein Face
    #Parameter 1: Surface-Objekt, in dem Attribute gespeichert sind
    #Parameter 2: Face-Objekt
    def set_face_attribute(surface, face)
      set_attribute(face, "Standard attribute", "PolygonID", surface.sSurfaceID) if(surface.sSurfaceID != "") #PolygonID
      set_attribute(face, "Standard attribute", "LinearRingID", surface.LinearRingID) if(surface.LinearRingID != "")  #LinearRingID
      if(surface.Attributes != nil) # Setzen der Standard Attribute
        surface.Attributes.each do |attr|
          set_attribute(face, "Standard attribute", attr[0], attr[1])
        end
      end
      #BoundarySurfaceType ID, suche mit Xlinkhref
      begin
        s = @surfacetypes.fetch(surface.xlinkhref)
        type = s[0]
        #puts s[0]
        set_attribute(face, "Standard attribute", "BoundarySurfaceType ID", s[1]) if( s[1] != "" and s[1] != nil)
      rescue =>e
        #BoundarySurfaceType ID, suche mit sSurfaceID
        begin
          s = @surfacetypes.fetch(surface.sSurfaceID)
          type = s[0]
          #puts s[0]
          set_attribute(face, "Standard attribute", "BoundarySurfaceType ID", s[1]) if( s[1] != "" and s[1] != nil)
        rescue =>e
          type = ""
        end
      end
      begin
        if(type == "")
          face.layer= "No Surfacetype"
        else
          face.layer= type
        end
      rescue => e
        puts e
        if(type == "")
          Sketchup.active_model.layers.add("No Surfacetype")
          face.layer= "No Surfacetype"
        else
          Sketchup.active_model.layers.add(type)
          face.layer= type
        end
      end
    end


    #6 Methoden, die benötigt werden, damit Plugin unter File -> Import erscheint
    #Der zurückgegebene String erscheint später im Dialog
    def description
      return "CityGML Files (*.xml, *.gml)"
    end

    #gibt String zurück, mit dem die Dateinamen gefiltert werden
    def file_extension
      # Returns the file extension of the file format your
      # importer can read (e.g. fbx, dae, etc.)
      return "*.gml;*.xml"
    end

    #Damit kann man Optionen beim Plugin aktivieren.
    #Wird aber nicht genutzt, sondern eigener Dialog
    def supports_options?
      # Returns true if your plugin supports import options.
      false
    end

    #Wird nicht genutzt
    def do_options
      # Presents the import options dialog panel to the user.
    end

    #Gibt City-GML als ID zurück
    def id
      # Returns the id string of your importer.
      return "City-GML"
    end

    #Methode wird beim drücken auf "Öffnen" aufgerufen
    #Übergabe von Dateiname, zweiter Parameter wird nicht genutzt
    def load_file(file_name, show_summary)
      # Reads the contents of location on disk specified
      # by file_name.
      # Shows the summary dialog panel if show_summary is true.
      # Returns 0 on success, returns 1 on failure.
      createModel(file_name)
      #createmodel(file_name)
      return 1
    end

    #Überprüft die Flächen auf planarität
    #Parameter 1: Das Surface, das geprüft werden soll
    #Parameter 2: Building ist Array von Surface/Points. Wenn Prüfung erfolgreich ist,
    #Parameter 3: surfacetype, wenn nicht vorhanden wird nil übergeben
    #wird Surface/Points in Array gespeichert
    def validate(surface, building, surfacetype, surfaceAttributes)
      surface.Vertices.delete_at(-1)
      pts = []
      coords = []
      return if(surface.Vertices.size < 3)
      surface.Vertices.each do |vertice|
        x = vertice.X.to_f - @minX
        y = vertice.Y.to_f - @minY
        z = vertice.Z.to_f - @minZ
        if(@switch)
          coords << [y, x, z]
        else
          coords << [x, y, z]
        end
      end
      coords.reverse! if(@switch)
      uniqcoords = coords.uniq
      uniqcoords.each { |coord|
        pts << Geom::Point3d.new(coord[0], coord[1], coord[2])
      }
      group = Sketchup.active_model.entities.add_group
      begin
        #Sollte add_face klappen, war Fläche planar. Anschließend wird es in buildings
        #gespeichert. 
        group.entities.add_face(pts)
        #Aufräumen
        group.erase!
        #Neben den 3dPoints wird auch das Surface gespeichert, weil dort noch
        #Informationen wie z.B. Attribute stecken, die noch benötigt werden
        building << [surface, pts]
        @coordlist["##{surface.sSurfaceID}"] = coords if(surface.sSurfaceID != nil)
        @validate += 1
        @surfacetypes[@validate] = [surfacetype, surfaceAttributes] if(surfacetype != nil)
        #pos = my_mesh.add_polygon(pts)
        #test_hash[pos] = surface
      rescue => e
        #Aufräumen
        group.erase!
      end
    end

    #Erzeugt alle Materialien und Texturen
    #Parameter 1: appearanceMember Objekt, in dem alle Materialen und Texturinformationen
    #vorhanden sind
    #Parameter 2: filename, gibt Pfad der XML-Datei an. Wird für Texturen benötigt
    #da im appearanceMember nur relativer Pfad vorhanden ist
    def create_all_materials(appearanceMember, filename)
      begin
        materials = Sketchup.active_model.materials
        #Iterationn über alle appearance
        appearanceMember.each do |app|
          #Iteration über alle surfaceDataMember
          app.SurfaceDataMembers.each do |surfaceDataMember|
            #Entweder ist material oder texture gesetzt
            material = surfaceDataMember.Material
            texture = surfaceDataMember.Texture
            if(material)
              #Überprüfen, ob Material vorhanden ist.
              if(materials["#{material.DiffuseColor[0]} #{material.DiffuseColor[1]} #{material.DiffuseColor[0]}"] != nil)
                m = materials["#{material.DiffuseColor[0]} #{material.DiffuseColor[1]} #{material.DiffuseColor[0]}"]
              else
                #Material anlegen, Name des Materials entspricht Farbwerte von Rot
                #Grün, Blau
                m = materials.add("#{material.DiffuseColor[0]} #{material.DiffuseColor[1]} #{material.DiffuseColor[0]}")
                color = Sketchup::Color.new()
                #Farbwerte skalieren
                color.red= material.DiffuseColor[0].to_f * 255
                color.green= material.DiffuseColor[1].to_f * 255
                color.blue= material.DiffuseColor[2].to_f * 255
                color.alpha= material.Transparency.to_f * 255
                m.color= color
              end
              #In den Targets stehen IDs der Flächen
              material.Targets.each do |target|
                #In der Hash-Map Material hinterlegen, Key ist die ID der Fläche
                @material_hash[target] = m if(@material_hash[target] == nil)
              end
            elsif(texture)
              #Texturname absoluten Pfad erzeugen
              texfilename = File.dirname(filename) + "\\" + texture.ImageURI
              #Manchmal fehlen die Texturen, dadurch wird die Fläche komplett
              #Schwarz. Deshalb überprüfen, ob Textur vorhanden ist
              if(File.exist?(texfilename))
                #Wurde die Textur bereits angelegt, kann diese verwendet werden
                if(materials[texture.ImageURI] == nil)
                  #Textur erzeugen
                  m = materials.add(texture.ImageURI)
                  m.texture = texfilename
                else
                  #Bereits angelegte verwenden
                  m = materials[texture.ImageURI]
                end
                #Texturkoordinaten werden später bei den einzelnen Flächen gesetzt.
                #Texturkoordinaten und Material wird jeweils in Hash-Map gespeichert
                texture.TargetTextureCoordinates.each_pair do |key, value|
                  @material_hash[key] = m
                  @texture_coordlist[key] = value
                  # puts key
                  # puts value
                end
              else
                # puts "The texture #{File.basename(texfilename)} was not found!"
              end
            end
          end
        end
      rescue => e
        puts e
        puts e.backtrace
      end
    end

    #Parst die CityGML-Datei und erzeugt daraus das Model
    def createModel(filename)
      JF::RubyToolbar::openConsole
      startzeit = Time.new
      debuglod = -1
      t = GMLParser.new()
      begin
        #Parst Model und gibt es als GML-Model zurück
        @gmlmodel = t.parse(filename, true)
        t = nil
      rescue => e
        #Bei Fehler abbrechen
        puts e
        puts e.backtrace
        t = nil
        return
      end
      @my_dialog = UI::WebDialog.new("CityGML Import", false, "Optionen #{rand 1000}", 285, 370, 0, 0, true)
      xmlname = File.basename(filename, ".xml")
      xmlname = xmlname.gsub(/.XML||.GML||.gml/, "")
      @my_dialog.set_html(GMLOptionsDialog.createDialog(@gmlmodel.LOD[0], @gmlmodel.LOD[1], @gmlmodel.LOD[2], @gmlmodel.LOD[3], xmlname))
      #Callback exit, wenn Benutzer Import abbricht
      @my_dialog.add_action_callback("exit") do |dialog, params|
        dialog.close
      end
      #Callbacl Move, zentriert Fenster in der Mitte des Monitors
      @my_dialog.add_action_callback("move") { |d, a|
        xy, wh = a.split(":")

        x, y = xy.split(",")
        x = x.to_i
        y = y.to_i

        w, h = wh.split(",")
        w = w.to_i
        h = h.to_i

        d.set_position((x - w)/2, (y - h)/2)
      }

      #Callback Choosen, wenn Auswahl getroffen wurde vom Benutzer
      #Startet Erzeugung des Models
      @my_dialog.add_action_callback("choosen") do |dialog, params|
        parameters = params.split(";")
        #puts parameters
        #Parameter 5 ist zum Tauschen von Northing und Easting
        #Dann müssen alle Texturkoordinaten und Koordinaten in der Reihenfolge getauscht werden
        #Und X und Y ebenfalls getauscht werden bei den Koordinaten
        if(parameters[5].to_f == 1)
          @switch = false
        else
          @switch = true
        end
        dialog.close
        @validate = 0
        @faces = 0
        @meshes = 0
        #Setzt das LOD
        debuglod = (parameters[0].to_i) -1
        if(debuglod == -1)
          puts "nix gefunden"
          return
        end
        if(JF::RubyToolbar::consoleOpen?)
          JF::RubyToolbar::clearConsole
        end
        #puts "Parsing time " + (Time.new.to_f() - startzeit.to_f()).to_s() + " Seconds"
        #UI.messagebox("1", MB_OK )
        #Überprüft, ob in den Model Attributen der SketchUp Datei bereits Offset
        #Werte gespeichert sind. Wenn dies der Fall ist, wurde bereits ein Model
        #geladen und Offset wird nicht erneut gesetzt, da sonst die neuen Gebäude
        #nicht im richtigen Abstand zu den alten sind.
        @minX = (Sketchup.active_model.get_attribute("Attribute", "Offset_X"))
        @minY = (Sketchup.active_model.get_attribute("Attribute", "Offset_Y"))
        @minZ = (Sketchup.active_model.get_attribute("Attribute", "Offset_Z"))
        @minX = @minX.to_f if(@minX != nil)
        @minY = @minY.to_f if(@minY != nil)
        @minZ = @minZ.to_f if(@minZ != nil)
        #Min X,Y,Z suchen
        #Iteration über alle CityObjects und  Generic3DObject. Dort auch
        #LODBoundedBy überprüfen, für alle Surfaces Minimum suchen
        #Wichtig, da die GeoKoordinaten viel zu Groß sind für SketchUp.
        #Minimum wird später abgezogen und so erscheint Model am Nullpunkt
        if(@minX == nil and @minY == nil and @minZ == nil)
          @gmlmodel.CityObjects.each do |cityObject|
            next if(cityObject == nil)
            if(cityObject.class == ReliefFeature)
              #Nicht implementiert
            elsif(cityObject.class == CityObjectGroup)
              cityObject.GroupMember.each do |groupmember|
                if(groupmember.GMLObject.LODXObject[debuglod] != nil)
                  groupmember.GMLObject.LODXObject[debuglod].each do |lodXObject|
                    lodXObject.Surfaces.each do |surface|
                      surface.Vertices.each do |vertice|
                        #puts vertice.X.to_s + " " + vertice.Y.to_s + " " + vertice.Z.to_s
                        findmin(vertice.X,vertice.Y,vertice.Z )
                      end
                    end
                  end
                end
                if(groupmember.GMLObject.LODBoundedBy[debuglod] != nil)
                  groupmember.GMLObject.LODBoundedBy[debuglod].each do |boundedby|
                    next if(boundedby == nil or boundedby.LodXObject == nil or boundedby.LodXObject.Surfaces.empty?)
                    boundedby.LodXObject.Surfaces.each do |surface|
                      next if(surface == nil)
                      surface.Vertices.each do |vertice|
                        #puts vertice.X.to_s + " " + vertice.Y.to_s + " " + vertice.Z.to_s
                        findmin(vertice.X,vertice.Y,vertice.Z )
                      end
                    end
                  end
                end
              end
            elsif(cityObject.class == Generic3DObject)
              if(cityObject.LODXObject[debuglod] != nil)
                cityObject.LODXObject[debuglod].each do |lodXObject|
                  lodXObject.Surfaces.each do |surface|
                    surface.Vertices.each do |vertice|
                      #puts vertice.X.to_s + " " + vertice.Y.to_s + " " + vertice.Z.to_s
                      findmin(vertice.X,vertice.Y,vertice.Z )
                    end
                  end
                end
              end
              if(cityObject.LODBoundedBy[debuglod] != nil)
                cityObject.LODBoundedBy[debuglod].each do |boundedby|
                  next if(boundedby == nil or boundedby.LodXObject == nil or boundedby.LodXObject.Surfaces.empty?)
                  boundedby.LodXObject.Surfaces.each do |surface|
                    next if(surface == nil)
                    surface.Vertices.each do |vertice|
                      findmin(vertice.X,vertice.Y,vertice.Z )
                    end
                  end
                end
              end
            end
          end
        end
        #Minimum im Model speichern, damit beim Export der Wert wieder addiert werden kann
        #puts "#{@minX} #{@minY} #{@minZ}"
        set_attribute(Sketchup.active_model, "Attribute", "Offset_X", @minX)
        set_attribute(Sketchup.active_model, "Attribute", "Offset_Y", @minY)
        set_attribute(Sketchup.active_model, "Attribute", "Offset_Z", @minZ)
        set_attribute(Sketchup.active_model, "Standard attribute", "import", true)
        #puts "Find min time " + (Time.new.to_f() - startzeit.to_f()).to_s() + " Seconds"
        begin
          #Für alle Surfaces den Surfacetype suchen (Wall, Roof, Ground)
          #@surfacetypes = FHGelsenkirchen::findSurfaceType(@gmlmodel.CityObjects)
          @surfacetypes = Hash.new
          model = Sketchup.active_model
          #Layer erzeugen, falls noch nicht vorhanden sind
          layers = model.layers
          layers.add "WallSurface"
          layers.add("RoofSurface")
          layers.add("GroundSurface")
          layers.add("No Surfacetype")
          #Wenn Parameter 3 true ist, wird neuer Layer erzeugt mit gewählten Namen(Parameter 4)
          if(parameters[2] == "true")
            layer = model.layers.add(parameters[3])
            model.active_layer= layer
          end
        rescue => e
          puts e
          puts e.backtrace
        end
        #HashMap für Materialen, Texturenkoordinaten und Koordinaten erzeugen
        @material_hash = Hash.new()
        @texture_coordlist = Hash.new()
        @coordlist = Hash.new()
        #Alle Materialen erzeugen, bevor Geometrie erzeugt wird
        create_all_materials(@gmlmodel.AppearanceMember, filename)
        begin
          #Neben dem Globalen AppearanceMember kann es zusätzlich noch in jedem
          #CityObject ein AppearanceMember geben, deshalb auch dort Materialien erzeugen
          @gmlmodel.CityObjects.each do |cityobject|
            if(cityobject.class == Generic3DObject)
              create_all_materials(cityobject.AppearanceMember, filename)
            elsif(cityobject.class == CityObjectGroup)
              cityobject.GroupMember.each do |groupmember|
                begin
                  create_all_materials(groupmember.GMLObject.AppearanceMember, filename) if(groupmember.GMLObject.AppearanceMember != nil)
                rescue => e
                  puts e
                  puts e.backtrace
                end
              end
            end
          end
        rescue => e
          puts e
          puts e.backtrace
        end
        #puts "Create Material time " + (Time.new.to_f() - startzeit.to_f()).to_s() + " Seconds"
        #Fertig mit Materialien erzeugen, jetzt kann Geometrie erzeugt werden
        @surfaces = []
        begin
          model = Sketchup.active_model
          layers = model.layers
          #Iteration über alle cityObjects
          @gmlmodel.CityObjects.each do |cityObject|
            cur_object = nil
            building = []
            next if(cityObject == nil)
            #ReliefFeature, bestehen aus Dreiecken
            if(cityObject.class == ReliefFeature)
              next
              cityObject.Triangles.each do |triangle|
                pts = []
                triangle.each do |t|
                  if(@switch)
                    pts << Geom::Point3d.new(t[1].to_f - @minX, t[0].to_f - @minY, t[2].to_f - @minZ)
                  else
                    pts << Geom::Point3d.new(t[0].to_f - @minX, t[1].to_f - @minY, t[2].to_f - @minZ)
                  end
                end
                group = Sketchup.active_model.entities.add_group
                begin
                  pts.reverse! if(@switch)
                  #Erzeugt das Face und löscht die temporäre Gruppe wieder
                  group.entities.add_face(pts)
                  group.erase!
                  building << [nil, pts]
                  @validate += 1
                rescue => e
                  group.erase!
                end
              end
              #Alle anderen Geometrien werden erst auf Planarität geprüft
              #Da es keinen richtigen Test dafür gibt, wird Fläche erzeugt
              #und überprüft, ob eine Exception auftritt.
              #Anschließend wird sie wieder gelöscht
            elsif(cityObject.class == CityObjectGroup)
                #Iteration über alle GroupMember
              cityObject.GroupMember.each do |groupmember|
                if(groupmember.GMLObject.LODXObject[debuglod] != nil)
                  cur_object = groupmember.GMLObject
                  #Iteration über alle LodXObjecte
                  groupmember.GMLObject.LODXObject[debuglod].each do |lodXObject|
                    lodXObject.Surfaces.each do |surface|
                      validate(surface, building, nil, nil)
                    end
                  end
                end
                #In manchen Modellen gibt es BoundedBy, dann steckt die Geometrie
                #im LodBoundedBy
                if(groupmember.GMLObject.LODBoundedBy[debuglod] != nil)
                  groupmember.GMLObject.LODBoundedBy[debuglod].each do |boundedby|
                    next if(boundedby == nil or boundedby.LodXObject == nil or boundedby.LodXObject.Surfaces.empty?)
                    boundedby.LodXObject.Surfaces.each do |surface|
                      next if(surface == nil)
                      validate(surface, building, boundedby.surfaceTyp, boundedby.surfaceAttributes)
                    end
                  end
                end
              end
              #Als letztes noch Generic3dObjects iterieren
            elsif(cityObject.class == Generic3DObject)
              cur_object = cityObject
              if(cityObject.LODXObject[debuglod] != nil)
                cityObject.LODXObject[debuglod].each do |lodXObject|
                  lodXObject.Surfaces.each do |surface|
                    validate(surface, building, nil, nil)
                  end
                end
              end
              #Auch hier gibt es möglicherweise BoundedBy
              if(cityObject.LODBoundedBy[debuglod] != nil)
                cityObject.LODBoundedBy[debuglod].each do |boundedby|
                  next if(boundedby == nil or boundedby.LodXObject == nil or boundedby.LodXObject.Surfaces.empty?)
                  boundedby.LodXObject.Surfaces.each do |surface|
                    next if(surface == nil)
                    validate(surface, building, boundedby.surfaceTyp,  boundedby.surfaceAttributes)
                  end
                end
              end
            end
            #Alle validierten Surfaces werden in diesem Array gespeichert
            @surfaces << [building, cur_object]
          end
        rescue => e
          puts e
        end
        #Validieren fertig. Geometrie wird jetzt erzeugt
        #puts "Validate time " + (Time.new.to_f() - startzeit.to_f()).to_s() + " Seconds"
        begin
          pos = 1
          #In Surfaces sind alle gültigen Polygone drin, sortiert nach Buildings
          @surfaces.each do |build, cur_object|
            #Geometrie wird erst als PolygonMesh erzeugt
            my_mesh = Geom::PolygonMesh.new
            build.each do |surface, pts|
              #Aus den Punkten werden Polygone gemacht
              my_mesh.add_polygon(pts)
              @meshes += 1
            end
            group = Sketchup.active_model.entities.add_group
            #Aus allen Mashes eines Buildings werden aufeinmal Polygone erzeugt
            group.entities.fill_from_mesh my_mesh, false, 0
            begin
              #Für Gruppe Attribute setzen, in der Regel Building-ID
              set_citygmlattributes(group, cur_object.Attributes, cur_object.genericAttributes) if(cur_object != nil)
            rescue => e
              puts e
              puts e.backtrace
            end
            count = 0
            #return
            a = Time.new.to_f()

            #ToDo: In diesem Code-Abschnitt gibt es wohl Fehler.
            #SketchUp soll eigentlich alle Faces in der selben Reihenfolge erzeugen,
            #wie sie vorher im Polygon-Mesh eingetragen wurden
            #Klappt leider nicht in 100% der Fälle
            #Bei Gebäuden mit vielen Flächen kommt es (warum auch) immer vor,
            #dass die Reihenfolge nicht mehr stimmt. Dann geht bei der
            #anschließenden Zuweisung von Attributen, Materialien und Texturen
            #das Modell kaputt
            group.entities.each do |entity|
              #Nur bei Faces etwas machen. In Entities sind ebenfalls noch die
              #Edges des Faces drin.
              if(entity.class == Sketchup::Face)
                @faces += 1
                surface = build[count][0]
                count += 1
                next if(surface == nil)
                #ID setzen
                set_attribute(entity, "Standard attribute", "PolygonID", surface.sSurfaceID) if(surface.sSurfaceID != "")
                set_attribute(entity, "Standard attribute", "LinearRingID", surface.LinearRingID) if(surface.LinearRingID != "")
                begin
                  #Material setzen, wenn vorhanden
                  entity.material = @material_hash.fetch("##{surface.sSurfaceID}")
                  if(entity.material.texture != nil) #Bei Texturen müssen noch Texturkoordinaten gesetzt werden 
                    txcoords = @texture_coordlist["##{surface.sSurfaceID}"]
                    coords = @coordlist["##{surface.sSurfaceID}"]
                    setTexture(entity, coords, txcoords)
                  end
                rescue => e
                  #puts e
                  begin
                    #puts surface.CompositeSurfaceId
                    entity.material = @material_hash.fetch("##{surface.CompositeSurfaceId}")
                  rescue => e
                    #puts @material_hash
                    #puts e
                  end
                end
                begin
                  #Zum Schluß noch noch Layer anpassen für das Face
                  s = @surfacetypes.fetch(pos)
                  pos += 1
                  entity.layer=s[0]
                  #BoundarySurfaceType ID setzen, sofern es eine gibt
                  set_attribute(entity, "Standard attribute", "BoundarySurfaceType ID", s[1]) if( s[1] != "" and s[1] != nil)
                rescue => e
                  entity.layer= "No Surfacetype"
                end
              end
            end
          end
        
        rescue => e
          puts e
          puts e.backtrace
        end
        #Import abgeschlossen
        puts "Total time " + (Time.new.to_f() - startzeit.to_f()).to_s() + " Seconds"
        puts "Faces: #{@faces}"
        #puts "Meshes: #{@meshes}"
        #puts "Validate: #{@validate}"
      end
      @my_dialog.show()
    end

    #Minimum suchen, X,Y,Z als Übergabeparameter
    def findmin(x, y, z)
      @minX = x.to_f() if (@minX == nil or x.to_f() < @minX)
      @minY = y.to_f() if (@minY == nil or y.to_f() < @minY)
      @minZ = z.to_f() if (@minZ == nil or z.to_f() < @minZ)
    end

    #Attribut auf einer Entity setzen.
    #Parameter 1: Entity, auf dem das Attribut gesetzt werden soll. Group oder Face
    #Parameter 2: Name des Dictionary.
    #Parameter 3: Key
    #Parameter 4: Value
    def set_attribute(entity, dict_name, key, value)
      value = "" if(value == nil)
      #Whitespaces etc. löschen
      entity.set_attribute(dict_name, key, value.to_s.gsub(/[\n]/, "").gsub(/[\\]/, "/").strip())
    end

    #Funktion zum Setzen von Setzen von allen Attributen
    #Parameter 1: Entity
    #Parameter 2: Array von Standard Attributen
    #Parameter 3: Array von Generischen Attributen
    def set_citygmlattributes(entity, attributes, genericAttributes)
      if(attributes != nil)
        attributes.each do |a1, a2|
          entity.set_attribute("Standard attribute", a1, a2)
        end
      end
      if(genericAttributes != nil)
        genericAttributes.each do |a1, a2|
          entity.set_attribute("Generic attribute", a1, a2)
        end
      end
    end
  end
end

#Sorgt dafür, dass das Plugin nicht in SketchUp 6 und ältere Versionen gestartet werden kann
version_required = 7.0  #Momentan werden noch keine Funktionen aus SketchUp 8 verwendet
begin
  if (Sketchup.version.to_f >= version_required)
    Sketchup.register_importer FHGelsenkirchen::Citygml.new #Sorgt dafür, dass Plugin unter File->Import erscheint
  end
rescue => e
  puts e
  puts e.backtrace
end