require 'CityGML/handler/SimpleGMLHandler'
require 'CityGML/objects/AppearanceMember'
module FHGelsenkirchen
  class AppearanceMemberHandler < SimpleGMLHandler
    def initialize(bSaveOtherTags)
      super(bSaveOtherTags)
      @isfirst = true
      @@EMPTYSURFACEDATA = 0
      @@TRANSPARENCY = 1
      @@DIFFUSECOLOR = 2
      @@TARGET = 4
      @@IMAGEURI = 8
      @@PARAMETERIZEDTEXTURE = 16
      @@X3DMATERIAL = 32
      @@TEXTURECOORDINATES = 64
      @@WRAPMODE = 64
	  @@GEOREFERENCEDTEXTURE = 128
      @CharBuffer = ""
      @curSurfaceData
      @bSaveSurfaceDataMemberInObjects = false
      @SurfaceDataBitMask = 0
      @curTextureTarget = ""
    end

    def resetHandler(newParent, newObject, bSaveSurfaceDataMemberInObjects)
      @bSaveSurfaceDataMemberInObjects = bSaveSurfaceDataMemberInObjects
      super(newParent, newObject)
      return nil
    end

    def on_start_element_ns(name, attributes, prefix, uri, namespaces)
      if(@bSaveSurfaceDataMemberInObjects)
        @ParentHandler.on_start_element_ns(name, attributes, prefix, uri, namespaces)
      end

      case(name)
      when 'surfaceDataMember':
          @isfirst = true
        #neues Objekt anlegen
        @curSurfaceData = AppearanceMember::SurfaceDataMember.new()
        #hinzufügen der SurfaceDataMember zu Model oder Parse-Objekt
        if (@bSaveSurfaceDataMemberInObjects)
          if (!(@curObject.Parent.class == Generic3DObject))
            raise GMLParserException,"Appearance muss zu einem Generic3DObject gehören"
          end
          @curObject.Parent.AppearanceMember[-1].SurfaceDataMembers << @curSurfaceData
        else
          @@Model.AppearanceMember[-1].SurfaceDataMembers << @curSurfaceData
        end
      when 'X3DMaterial':
          @isfirst = true
        @curSurfaceData.Material = AppearanceMember::SurfaceDataMember::X3DMaterial.new()
        @SurfaceDataBitMask |= @@X3DMATERIAL
	  when 'GeoreferencedTexture':
	            @isfirst = true
        @curSurfaceData.Texture = AppearanceMember::SurfaceDataMember::ParameterizedTexture.new();
        @SurfaceDataBitMask |= @@GEOREFERENCEDTEXTURE
      when 'ParameterizedTexture':
          @isfirst = true
        @curSurfaceData.Texture = AppearanceMember::SurfaceDataMember::ParameterizedTexture.new();
        @SurfaceDataBitMask |= @@PARAMETERIZEDTEXTURE
      when 'transparency'
        @isfirst = true
        @SurfaceDataBitMask |= @@TRANSPARENCY
      when 'diffuseColor'
        @isfirst = true
        @SurfaceDataBitMask |= @@DIFFUSECOLOR
      when 'imageURI'
        @isfirst = true
        @SurfaceDataBitMask |= @@IMAGEURI
      when 'wrapMode'
        @isfirst = true
        @SurfaceDataBitMask |= @@WRAPMODE
      when 'target'
        @isfirst = true
        @SurfaceDataBitMask |= @@TARGET
        if((@SurfaceDataBitMask & @@PARAMETERIZEDTEXTURE) != 0)
          if (attributes.length > 0)
            @curTextureTarget = attributes["uri"]
          else
            raise Exception, "target has no uri"
          end
        end
      when 'textureCoordinates'
        @SurfaceDataBitMask |= @@TEXTURECOORDINATES
      end
      return super(name, attributes, prefix, uri, namespaces)
    end

    def on_end_element_ns(name, prefix, uri)
      qname = FHGelsenkirchen::createQname(prefix, name)
      if (@bSaveSurfaceDataMemberInObjects)
        @ParentHandler.on_end_element_ns(name, prefix, uri)
      end
      if (qname == @curObject.QName)
        SavetCurrentPostObjects()
        @ParentHandler.OnChildParsed(self)
        if (!@SaveOtherTags)
          @curObject.PreObjects = nil;
          @curObject.PostObjects = nil;
        end
        @curObject = nil
        return true
      end
      #floatpattern = Regexp.new('[0-9]+[eE][-+]?[0-9]+|[-+]?[0-9]*\.?[0-9]+')
      bClearCharBuffer = false
      value = ""
      mc = []
      case(name)
      when "surfaceDataMember":
          @curSurfaceData = nil;    #zur Sicherheit
      when "X3DMaterial":
          @SurfaceDataBitMask &= ~@@X3DMATERIAL
	  when 'GeoreferencedTexture':
	    @SurfaceDataBitMask &= ~@@GEOREFERENCEDTEXTURE
      when "ParameterizedTexture":
          @SurfaceDataBitMask &= ~@@PARAMETERIZEDTEXTURE
      when "transparency":
          @SurfaceDataBitMask &= ~@@TRANSPARENCY
        @curSurfaceData.Material.Transparency = @CharBuffer  #ToDo: Hier muss Dezimal rein
        bClearCharBuffer = true
      when "diffuseColor":
          @SurfaceDataBitMask &= ~@@DIFFUSECOLOR
        value = @CharBuffer
        #mc = value.scan(floatpattern)
        mc = value.split(' ')
        @curSurfaceData.Material.DiffuseColor[0] = mc[0]  #ToDo: wie war Single
        @curSurfaceData.Material.DiffuseColor[1] = mc[1]
        @curSurfaceData.Material.DiffuseColor[2] = mc[2]
        bClearCharBuffer = true
      when "imageURI":
	   # if(@curSurfaceData.Texture != nil)
          @SurfaceDataBitMask &= ~@@IMAGEURI
          @curSurfaceData.Texture.ImageURI = @CharBuffer.strip
          bClearCharBuffer = true
		#  end
      when "wrapMode":
          @SurfaceDataBitMask &= ~@@WRAPMODE;
        @curSurfaceData.WrapMode = @CharBuffer
        bClearCharBuffer = true
      when "target":
          @SurfaceDataBitMask &= ~@@TARGET;
        if ((@SurfaceDataBitMask & @@X3DMATERIAL) != 0)
          @curSurfaceData.Material.Targets << @CharBuffer
        elsif ((@SurfaceDataBitMask & @@PARAMETERIZEDTEXTURE) != 0)
          #nichts, target wurde oben aus Attribut entnommen
		  elsif((@SurfaceDataBitMask & @@GEOREFERENCEDTEXTURE) != 0)
        else
          raise GMLParserException, "found \"target\" outside (\"X3DMaterial\" or \"ParameterizedTexture\")"
        end
        @curTextureTarget = nil
        bClearCharBuffer = true
      when "textureCoordinates":
          if(@isfirst)
          @isfirst = false
          @SurfaceDataBitMask &= ~@@TEXTURECOORDINATES
          value = @CharBuffer
          mc = value.split(' ')
          #mc = value.scan(floatpattern)
          if (mc.length % 2 != 0)
            raise GMLParserException, "Texturcoordinate failure: count is not even"
          end
          textureCoordinates = []
          for i in 0..((mc.length() / 2)- 1 )
            s_t1 = mc[2 * i]
            t1 = s_t1.to_f()
            s_t2 = mc[2 * i + 1]
            t2 = s_t2.to_f()
            vtc = [t1, t2]
            textureCoordinates << vtc
          end
          @curSurfaceData.Texture.TargetTextureCoordinates[@curTextureTarget]= textureCoordinates
          bClearCharBuffer = true
        end
      else
        bClearCharBuffer = false
      end
      if (bClearCharBuffer)
        @CharBuffer= ""
      end
      return false
    end

    def  on_characters(chars)
      if (@bSaveSurfaceDataMemberInObjects)
        #in diesem modus werden die AppearanceMember mit in die PreObjects des Parents geparst, da die GML Datenstrukturen nicht komplett nachgebildet sind
        @ParentHandler.on_characters(chars)
      end
      #einzelne Informationen gesondert speichern
      if (((@SurfaceDataBitMask & @@TRANSPARENCY) != 0) || #Es wird gerade ein Texturname geparst
          ((@SurfaceDataBitMask & @@DIFFUSECOLOR) != 0) || #Es wird gerade ein Texturname geparst
          ((@SurfaceDataBitMask & @@IMAGEURI) != 0) || #Es wird gerade ein Texturname geparst
          ((@SurfaceDataBitMask & @@TARGET) != 0) || #Es wird gerade ein Ziel geparst
          ((@SurfaceDataBitMask & @@TEXTURECOORDINATES) != 0))       #Es werden gerade Texturkoordinaten geparst
        @CharBuffer << chars
      end
      #alle Informationen jedoch immer wie gewohnt speichern
      super(chars)
      return nil
    end

  end
end