require 'CityGML/Handler/SimpleGMLHandler'
require 'CityGML/Handler/Generic3DObjectHandler'
require 'CityGML/Handler/CityObjectGroupHandler'
require 'CityGML/Handler/ReliefFeatureHandler'
require 'CityGML/Objects/Generic3DObject'
require 'CityGML/Objects/ReliefFeature'
require 'CityGML/Objects/CityObjectGroup'

# Dies ist ein Handler für die abstrakten cityObjectMember
# Er dient lediglich dazu, den nächsten Handler für die cityObjectMember-Instanz zu aktivieren.
# Ist der cityObjectMember nicht implementiert, wird eine Warning generiert und das Object ignoriert.
module FHGelsenkirchen
  class CityObjectMemberHandler < SimpleGMLHandler
    def initialize(bSaveOtherTags)
      super(bSaveOtherTags)
      @Generic3DObjectHandler = Generic3DObjectHandler.new(bSaveOtherTags)
      @CityObjectGroupHandler = CityObjectGroupHandler.new(bSaveOtherTags)
      @ReliefFeatureHandler = ReliefFeatureHandler.new(bSaveOtherTags)
      @curObject = AbstractGMLDummy.new(@@Model, "cityObjectMember", nil, bSaveOtherTags);
      @bIgnoring = false
      @IgnoredQName = ""
    end

    def on_start_element_ns(name, attributes, prefix, uri, namespaces)
      found = false
      qname = FHGelsenkirchen::createQname(prefix, name)
      @Generic3DObjectHandler.Implemented3DObjects().each do |o|
        found = true if o == name
      end
      if (found) #implementiert der Generic3DObjectHandler dieses Element?
        f = 0
        case (name)
        when "Building" then f = 1
        when "SolitaryVegetationObject": f = 1
        when "GenericCityObject": f = 1
        else
          raise Exception, qname + " ist falsch implementiert"
        end
        if (f==1)
          #Keine PreObjects vorhanden
          #neues Objekt anlegen
          new3DObject = Generic3DObject.new(@curObject, qname, attributes, @bSaveOtherTags);
          #new3DObject.DebugLine = GMLParser.SaxGMLHandler.DebugXMLLine
          SimpleGMLHandler.getModel.CityObjects << new3DObject
          #Handler reset
          @Generic3DObjectHandler.resetHandler(self, new3DObject)
          return @Generic3DObjectHandler
        end
      elsif(name == "CityObjectGroup") #ist es eine cityObjectGroup?
        #Keine PreObjects vorhanden
        #neues Objekt anlegen
        newCityObjectGroup = CityObjectGroup.new(@curObject, qname, attributes, @bSaveOtherTags)
        #newCityObjectGroup.DebugLine = GMLParser.SaxGMLHandler.DebugXMLLine
        @@Model.CityObjects << newCityObjectGroup
        #Handler reset
        @CityObjectGroupHandler.resetHandler(self, newCityObjectGroup)
        return @CityObjectGroupHandler
      elsif (name == "ReliefFeature")
        #neues Objekt anlegen
        newReliefFeatureHandler = ReliefFeature.new(@curObject, qname, attributes, @bSaveOtherTags)
        #newReliefFeatureHandler.DebugLine = GMLParser.SaxGMLHandler.DebugXMLLine
        @@Model.CityObjects << (newReliefFeatureHandler)
        #Handler reset
        @ReliefFeatureHandler.resetHandler(self, newReliefFeatureHandler)
        return @ReliefFeatureHandler
      else    #sonst ist dieses Element noch nicht implementiert -> Warnung generieren
        if (!@bIgnoring)
          puts "Warning: cityObjectMember \"#{qname}\" was ignored"
          @IgnoredQName = qname
        end
        @bIgnoring = true
        return self
      end
    end

    def on_end_element_ns(name, prefix, uri)
      qname = FHGelsenkirchen::createQname(prefix, name)
      if (!@bIgnoring)   #normaler fall
        return true
      else    #zur Zeit am ignorieren
        if (qname == @IgnoredQName)
          @bIgnoring = false
          @IgnoredQName = nil
          return false
        end
      end
    end

    def on_characters(chars)
      if(!@bIgnoring)
        super(chars)
      end
    end

  end
end