require 'CityGML/Handler/SimpleGMLHandler'
require 'CityGML/Objects/LodXObject'
require 'CityGML/Objects/BoundedBy'
require 'CityGML/Handler/AppearanceMemberHandler'
require 'CityGML/Handler/BoundedByHandler'
require 'CityGML/Handler/LodXHandler'

module FHGelsenkirchen
  class Generic3DObjectHandler < SimpleGMLHandler
    # Das aktuele Parse-Object
    attr_accessor :Cur3DObject
    def Implemented3DObjects()
      return @@Implemented3DObjects
    end
    def getCur3DObject()
      return @curObject
    end
    # Gibt an, welche Typen zu Generic3DObject gehören.
    def cleanvalue(chars)
      if(chars.lstrip == "")
        return ""
      end
      return chars
    end

    def initialize(bSaveOtherTags)
      super(bSaveOtherTags)
      @AppearanceMemberHandler = AppearanceMemberHandler.new(bSaveOtherTags)
      @BoundedByHandler = BoundedByHandler.new(bSaveOtherTags)
      @LodXHandler = LodXHandler.new(bSaveOtherTags)
      @@Implemented3DObjects = []
      @@Implemented3DObjects << "Building"
      @@Implemented3DObjects << "SolitaryVegetationObject"
      @@Implemented3DObjects <<"GenericCityObject"
      @AttributeBitMask = 0
      @@MEASUREDHEIGHT = 1
      @@STOREYSABOVEGROUND = 2
      @@STOREYSBELOWGROUND = 4
      @@GENERICATTRIBUTE = 8
      @@VALUE = 16
      @@YEAROFCONSTURCTION = 32
      @@CLASS = 64
      @@FUNCTION = 128
      @@USAGE = 256
      @@YEAROFDEMOLITION = 512
      @@STOREYHEIGHTSABOVEGROUND = 1024
      @@STOREYHEIGHTSBELOWGROUND = 2048
    end

    def on_start_element_ns(name, attributes, prefix, uri, namespaces)
      f = 0
      qname = FHGelsenkirchen::createQname(prefix, name)
      case (name)
      when "lod1Solid" then f = 1
      when "lod1MultiSurface" then f = 1
      when "lod2Solid" then f = 1
      when "lod2MultiSurface" then f = 1
      when "lod3Solid" then f = 1
      when "lod3MultiSurface" then f = 1
      when "lod4Solid" then f = 1
      when "lod4MultiSurface" then f = 1
      when "lod1Geometry" then f = 1
      when "lod2Geometry" then f = 1
      when "lod3Geometry" then f = 1
      when "lod4Geometry" then f = 1
      when "boundedBy":
          #neues Objekt anlegen
        newBoundedBy = BoundedBy.new(@curObject, qname, attributes, @bSaveOtherTags);
        #Handler reset
        @BoundedByHandler.resetHandler(self, newBoundedBy);
        return @BoundedByHandler
      when "appearance" then f = 2
      when "appearanceMember" then f = 2
      when "measuredHeight"
        @AttributeBitMask |= @@MEASUREDHEIGHT
        return super(name, attributes, prefix, uri, namespaces)
      when "storeysAboveGround"
        @AttributeBitMask |= @@STOREYSABOVEGROUND
        return super(name, attributes, prefix, uri, namespaces)
      when "storeysBelowGround"
        @AttributeBitMask |= @@STOREYSBELOWGROUND
        return super(name, attributes, prefix, uri, namespaces)
      when "_genericAttribute"
        @AttributeBitMask |= @@GENERICATTRIBUTE
        return super(name, attributes, prefix, uri, namespaces)
      when "genericAttribute>"
        @AttributeBitMask |= @@GENERICATTRIBUTE
        return super(name, attributes, prefix, uri, namespaces)
      when "value"
        @AttributeBitMask |= @@VALUE
        return super(name, attributes, prefix, uri, namespaces)
      when "class"
        @AttributeBitMask |= @@CLASS
        return super(name, attributes, prefix, uri, namespaces)
      when "function"
        @AttributeBitMask |= @@FUNCTION
        return super(name, attributes, prefix, uri, namespaces)
      when "usage"
        @AttributeBitMask |= @@USAGE
        return super(name, attributes, prefix, uri, namespaces)
      when "yearOfConstruction"
        @AttributeBitMask |= @@YEAROFCONSTURCTION
        return super(name, attributes, prefix, uri, namespaces)
      when "yearOfDemolition"
        @AttributeBitMask |= @@YEAROFDEMOLITION
        return super(name, attributes, prefix, uri, namespaces)
      when "storeyHeightsAboveGround"
        @AttributeBitMask |= @@STOREYHEIGHTSABOVEGROUND
        return super(name, attributes, prefix, uri, namespaces)
      when "storeyHeightsBelowGround"
        @AttributeBitMask |= @@STOREYHEIGHTSBELOWGROUND
        return super(name, attributes, prefix, uri, namespaces)
      when "doubleAttribute"
        @AttributeBitMask |= @@GENERICATTRIBUTE
        @name = attributes["name"]
        return super(name, attributes, prefix, uri, namespaces)
      when "intAttribute"
        @AttributeBitMask |= @@GENERICATTRIBUTE
        @name = attributes["name"]
        return super(name, attributes, prefix, uri, namespaces)
      when "stringAttribute"
        @AttributeBitMask |= @@GENERICATTRIBUTE
        @name = attributes["name"]
        return super(name, attributes, prefix, uri, namespaces)
      else
        return super(name, attributes, prefix, uri, namespaces)
      end
      if(f == 1)
        #Eventuelle PreObjects speichern
        if(@curObject.PreObjects == nil || @curObject.PreObjects.empty? )
          SaveCurrentPreObjects()
        end
        #neues Objekt anlegen

        curLOD = t = (name.delete("lod").to_i) -1 #gespeichert werden direkt ArrayIndices
        @@Model.LOD[curLOD] = true #Speicher welche LODs vorkommen
        newLodXObject = LodXObject.new(@curObject, qname, attributes, @bSaveOtherTags)
        if (getCur3DObject().LODXObject[curLOD] == nil)
          getCur3DObject().LODXObject[curLOD] = []
        end
        getCur3DObject().LODXObject[curLOD] << newLodXObject
        #Handler reset
        @LodXHandler.resetHandler(self, newLodXObject)
        return @LodXHandler
      elsif(f == 2)
        #neues Objekt anlegen
        getCur3DObject().AppearanceMember << AppearanceMember.new(@curObject, qname, attributes, @bSaveOtherTags)
        @AppearanceMemberHandler.resetHandler(self, getCur3DObject().AppearanceMember[-1], true)
        return @AppearanceMemberHandler
      end

    end

    def on_end_element_ns(name, prefix, uri)
      if((@AttributeBitMask & @@MEASUREDHEIGHT) != 0)
        @AttributeBitMask &= ~@@MEASUREDHEIGHT
      elsif((@AttributeBitMask & @@STOREYSABOVEGROUND) != 0)
        @AttributeBitMask &= ~@@STOREYSABOVEGROUND
      elsif((@AttributeBitMask & @@STOREYSBELOWGROUND) != 0)
        @AttributeBitMask &= ~@@STOREYSBELOWGROUND
      elsif((@AttributeBitMask & @@GENERICATTRIBUTE) != 0)
        @AttributeBitMask &= ~@@GENERICATTRIBUTE
      elsif((@AttributeBitMask & @@CLASS) != 0)
        @AttributeBitMask &= ~@@CLASS
      elsif((@AttributeBitMask & @@FUNCTION) != 0)
        @AttributeBitMask &= ~@@FUNCTION
      elsif((@AttributeBitMask & @@USAGE) != 0)
        @AttributeBitMask &= ~@@USAGE
      elsif((@AttributeBitMask & @@YEAROFCONSTURCTION) != 0)
        @AttributeBitMask &= ~@@YEAROFCONSTURCTION
      elsif((@AttributeBitMask & @@YEAROFDEMOLITION) != 0)
        @AttributeBitMask &= ~@@YEAROFDEMOLITION
      elsif((@AttributeBitMask & @@STOREYHEIGHTSABOVEGROUND) != 0)
        @AttributeBitMask &= ~@@STOREYHEIGHTSABOVEGROUND
      elsif((@AttributeBitMask & @@STOREYHEIGHTSBELOWGROUND) != 0)
        @AttributeBitMask &= ~@@STOREYHEIGHTSBELOWGROUND
      end
      return super(name, prefix, uri)
    end

    def on_characters(chars)
      # puts chars
      if((@AttributeBitMask & @@MEASUREDHEIGHT) != 0)
        @curObject.Attributes << ["measuredHeight", cleanvalue(chars)]
        @AttributeBitMask &= ~@@MEASUREDHEIGHT
      elsif((@AttributeBitMask & @@STOREYSABOVEGROUND) != 0)
        @curObject.Attributes << ["storeysAboveGround", cleanvalue(chars)]
        @AttributeBitMask &= ~@@STOREYSABOVEGROUND
      elsif((@AttributeBitMask & @@STOREYSBELOWGROUND) != 0)
        @curObject.Attributes << ["storeysBelowGround", cleanvalue(chars)]
        @AttributeBitMask &= ~@@STOREYSBELOWGROUND
      elsif(((@AttributeBitMask & @@VALUE) != 0) and ((@AttributeBitMask & @@GENERICATTRIBUTE) != 0))
        @curObject.genericAttributes[@name] = cleanvalue(chars)
        @AttributeBitMask &= ~@@VALUE
      elsif((@AttributeBitMask & @@CLASS) != 0)
        @AttributeBitMask &= ~@@CLASS
        @curObject.Attributes << ["class", cleanvalue(chars)]
      elsif((@AttributeBitMask & @@FUNCTION) != 0)
        @curObject.Attributes << ["function" , cleanvalue(chars)]
        @AttributeBitMask &= ~@@FUNCTION
      elsif((@AttributeBitMask & @@USAGE) != 0)
        @curObject.Attributes << ["usage", cleanvalue(chars)]
        @AttributeBitMask &= ~@@USAGE
      elsif((@AttributeBitMask & @@YEAROFCONSTURCTION) != 0)
        @curObject.Attributes << ["yearOfConstruction", cleanvalue(chars)]
        @AttributeBitMask &= ~@@YEAROFCONSTURCTION
      elsif((@AttributeBitMask & @@YEAROFDEMOLITION) != 0)
        @curObject.Attributes << ["yearOfDemolition", cleanvalue(chars)]
        @AttributeBitMask &= ~@@YEAROFDEMOLITION
      elsif((@AttributeBitMask & @@STOREYHEIGHTSABOVEGROUND) != 0)
        @curObject.Attributes << ["storeyHeightsAboveGround", cleanvalue(chars)]
        @AttributeBitMask &= ~@@STOREYHEIGHTSABOVEGROUND
      elsif((@AttributeBitMask & @@STOREYHEIGHTSBELOWGROUND) != 0)
        @curObject.Attributes << ["storeyHeightsBelowGround", cleanvalue(chars)]
        @AttributeBitMask &= ~@@STOREYHEIGHTSBELOWGROUND
      end
    end

    def OnChildParsed(child)
      if(child.class == BoundedByHandler)
        boundedBy = @BoundedByHandler.BoundedBy
        if (boundedBy.LodXObject != nil) #das BoundedBy enthält ein LodXObject
          lod = boundedBy.ObjectLOD
          if (@curObject.LODBoundedBy[lod] == nil)
            @curObject.LODBoundedBy[lod] = []
          end
          @curObject.LODBoundedBy[lod] << boundedBy
        else    #falls nein, muss der geparste text dem XMLBuffer angefügt werden
          #"irgendetwas" wurde in jedem Fall geparst
          @XMLBuffer << boundedBy.PostObjects
        end
      end
    end
  end
end