require 'CityGML/Helper'
#Oberklasse für alle Handler
module FHGelsenkirchen
  class SimpleGMLHandler
    # Beim Parsen wird dieses Model befüllt
    @@Model = nil
    def self.getModel()
      return @@Model
    end
    def self.setModel(model)
      @@Model = model
    end
    # true: Pre/Post-Objects bleiben noch nach Ende des aktuellen Elements erhalten
    @bSaveOtherTags = false
    # Dieses Objects befindet sich gerade zum Parsen im Handler
    attr_reader :curObject
    # Handler, welcher diesen Handler aktiviert hat
    attr_reader :ParentHandler
    # Speichert die Pre/Post-Objects
    attr_reader :XMLBuffer
    def initialize( bSaveOtherTags)
      @XMLBuffer = ""
      @bSaveOtherTags = bSaveOtherTags
    end
    # Muss mit jedem aktivieren gerufen werden. Bereitet den Handler für ein neues Parse-Objekt vor
    #<param name="newParent">Der aktivierende Handler. Zu diesem wird zurückgesprungen wenn das Parse-Objekt fertig ist</param>
    #<param name="newObject">Das neue zu parsende Objekt (Parse-Objekt)</param>
    def resetHandler(newParent, newObject)
      @ParentHandler = newParent;
      @curObject = newObject;
      @XMLBuffer= ""
    end

    # Speichert den aktuellen XMLBuffer als PreObjects und leert diesen anschließend
    def SaveCurrentPreObjects()
      @curObject.PreObjects = @XMLBuffer
      @XMLBuffer = ""
    end

    # Speichert den aktuellen XMLBuffer als PostObjects und leert diesen anschließend
    def SavetCurrentPostObjects()
      @curObject.PostObjects = @XMLBuffer
      @XMLBuffer = ""
    end

    #Speichert das Element samt seinen Attributen im XMLBuffer
    # <param name="uri"></param>
    # <param name="localname"></param>
    # <param name="qname"></param>
    # <param name="attrs"></param>
    # <returns>this (den neuen Handler für das nächste Element)</returns>
    def on_start_element_ns(name, attributes, prefix, uri, namespaces)
      #begin
      qname = FHGelsenkirchen::createQname(prefix, name)
      if (qname != @curObject.QName)
        @XMLBuffer << "<#{qname}"
        s = ""
        s = SimpleGMLHandler::saveAttributes(attributes)
        if (s != nil)
          s.each do |attribs|
            @XMLBuffer <<" #{attribs[0]}=\"#{attribs[1]}\""
          end
        end
        @XMLBuffer << ">"
      end
      return self
    rescue =>e
      puts e.message
      #end
    end

    # macht aus attrs einen XML-Attribute-String
    # <param name="attrs">zu speichernde Attribute</param>
    # <returns></returns>
    def self.saveAttributes(attrs)
      if (attrs == nil || attrs.length == 0)
        return []
      end
      attribs = []
      attrs.each do |key, value|
        attribs << [key.to_s(), value.to_s()]
      end
      return attribs
    end


    # Fügt das Element zum XMLBuffer, falls es nicht das Ende des Parse-Objektes ist.
    # Sonst werden PostObjects gespeichert und OnChildParsed des ParentHandler gerufen.
    # Je nach Pre/PostObject-Schalter werden diese gelöscht
    # <param name="uri"></param>
    # <param name="localname"></param>
    # <param name="qname"></param>
    # <returns>true, wenn das Element das schließende des Parse-Objekts war.</returns>
    def on_end_element_ns(name, prefix, uri)
      qname = FHGelsenkirchen::createQname(prefix, name)
      #Ende des Elements behandeln
      if (qname == @curObject.QName) #das aktuelle Element wurde geschlossen
        SavetCurrentPostObjects()
        #Rufe OnChildParsed des parentHandler
        @ParentHandler.OnChildParsed(self)
        #if (@bSaveOtherTags) #Pre/PostObjects ggf. löschen
        #  @curObject.PreObjects = nil
        #  @curObject.PostObjects = nil
        #end
        @curObject = nil #Das Element ist jetzt fertig
        return true
      else #zum pre/postBuffer hinzufügen
        @XMLBuffer << "</#{qname}>"
        return false
      end
    end

    #speichert die Charakter im XMLBuffer
    def on_characters(chars)
      @XMLBuffer << chars
    end

    #Callback Methode die gerufen wird, wenn das Parse-Objekt fertig ist.
    # Die Pre-/PostObjects sind jetzt im Parse-Objekts von child verfügbar und so
    # können dort noch Informationen child.curObject.PreObjects bzw child.curObject.PostObjects
    # geholt werden
    # <param name="child">der Handler welcher die Callbackmethode ruft</param>
    def OnChildParsed(child)
      #Muss implementiert werden
      #raise Exception,"Nicht implementiert"
    end

  end
end