require 'CityGML/Handler/SimpleGMLHandler'
require 'CityGML/Helper'
require 'CityGML/Objects/SurfaceMember'
module FHGelsenkirchen
  class SurfaceMemberHandler < SimpleGMLHandler
    attr_accessor :Surface
    attr_accessor :curObject

    def initialize(bSaveOtherTags)
      super(bSaveOtherTags)
      @EMPTYSURFACE = 0
      @POS = 1
      @POSLIST = 2
      @TEXTUREMAP = 4
      @TEXTURECOORDINATES = 8
      @SurfaceBitMask = @EMPTYSURFACE
      @CharBuffer = ""
    end

    def on_start_element_ns(name, attributes, prefix, uri, namespaces)
      bSaveThisElement = true
      f = 0
      case (name)
      when "surfaceMember":
          if (attributes.length > 0 && attributes["href"])
          @curObject.Xlinkhref = attributes["href"]
        end
      when "TexturedSurface" then f = 1
      when "Polygon" then f = 1
      when "LinearRing":
          begin
          if (attributes.length > 0 && attributes["id"])
            @curObject.LinearRingID = attributes["id"]
          end
        rescue =>e
        end
      when "interior":
          if(@curObject.VerticesInterior == nil)
          @curObject.VerticesInterior = []
        end
        @curObject.VerticesInterior << []
      when "pos":
          if (@curObject.PreObjects == nil || @curObject.PreObjects.empty?)
          SaveCurrentPreObjects()
        end
        @SurfaceBitMask |= @POS
        bSaveThisElement = true
      when "posList":
          if (@curObject.PreObjects == nil || @curObject.PreObjects.empty?)
          SaveCurrentPreObjects()
        end
        @SurfaceBitMask |= @POSLIST
        bSaveThisElement = false
      when "appearance":
          @curObject.Appearance = GMLAppearance.new()
      when "textureMap":
          @SurfaceBitMask |= @TEXTUREMAP
      when "textureCoordinates":
          @SurfaceBitMask |= @TEXTURECOORDINATES
      when "CompositeSurface"
        begin
          @curObject.CompositeSurfaceId = attributes.fetch("id")
        rescue => e
        end
      end
      if(f==1)
        begin
          if (attributes.length > 0 && attributes["id"]!="")
            if (@curObject.sSurfaceID != "")
              raise GMLParserException,"Surface already has an ID. Old: " + @curObject.sSurfaceID +
                "New: " + attributes["id"]
            end
            @curObject.sSurfaceID = attributes["id"]
          end
        rescue =>e
        end
      end
      if (bSaveThisElement)
        return super(name, attributes, prefix, uri, namespaces)
      end
      return self
    end

    def on_end_element_ns(name, prefix, uri)
      #floatpattern = Regexp.new("[0-9]+[eE][-+]?[0-9]+|[-+]?[0-9]*\.?[0-9]+|[-+]?[0-9]+\.?[0-9]+[eE][-][0-9]+") # Regulärer Ausdruck für float Zahlen
      bSaveThisElement = true
      case (name)
      when "pos":
          @SurfaceBitMask &= ~@POS
        if (@curObject.VerticesInterior.empty?)    #dies sind exterior vertices, denn diese kommen in GML zuerst
          bSaveThisElement = true
          posValue = @CharBuffer
          tmp = GMLVertex.new()
          tmp.createbyString(posValue)
          @curObject.Vertices << tmp
          @CharBuffer = ""
        else
          #exterior kamen davor
          posValue = @CharBuffer
          tmp = GMLVertex.new()
          tmp.createbyString(posValue)
          @curObject.VerticesInterior[@curObject.VerticesInterior.length() - 1] << tmp  #im letzten Interior einfügen
          @CharBuffer = ""
          @SaveThisElement = true
        end
      when "posList":
          @SurfaceBitMask &= ~@POSLIST
        if (@curObject.VerticesInterior.empty?)    #dies sind exterior vertices, denn diese kommen in GML zuerst
          bSaveThisElement = false
          posListValue = @CharBuffer
          @CharBuffer = ""
          #mcPL = posListValue.scan(floatpattern)
          mcPL = posListValue.split(' ')
          if (mcPL.length%3 != 0)
            raise GMLParserException,"posList failure: pos count is not a power of 3"
          end
          for i in 0..(mcPL.length/3)-1
            tmp = GMLVertex.new()
            tmp.createbyfloat(mcPL[3*i], mcPL[3*i + 1], mcPL[3*i + 2])
            @curObject.Vertices << tmp
          end
        else
          posListValue = @CharBuffer
          @CharBuffer = ""
          #mcPL = posListValue.scan(floatpattern)
          mcPL = posListValue.split(' ')
          if (mcPL.length % 3 != 0)
            raise GMLParserException, "posList failure: pos count is not a power of 3"
          end
          for j in 0..mcPL.length / 3
            tmp = GMLVertex.new()
            tmp.createbyfloat(mcPL[3 * j].to_f, mcPL[3 * j + 1].to_f, mcPL[3 * j + 2].to_f)
            @curObject.VerticesInterior[@curObject.VerticesInterior.length - 1] << tmp #im letzten Interior einfügen
          end
          bSaveThisElement = true
        end
      when "textureMap":
          @SurfaceBitMask &= ~@TEXTUREMAP;
        @curObject.Appearance.sTexture = @CharBuffer
        @CharBuffer = ""
      when "textureCoordinates":
          @SurfaceBitMask &= ~@TEXTURECOORDINATES;
        textureCoordinatesValue = @CharBuffer
        @CharBuffer = ""
        #mcTC = textureCoordinatesValue.scan(floatpattern)
        mcTC = textureCoordinatesValue.split(' ')
        if (mcTC.length % 2 != 0)
          raise GMLParserException,"Texturcoordinate failure: count is not even"
        end
        for k in 0..(mcTC.length / 2 - 1)
          s_t1 = mcTC[2 * k]
          t1 = s_t1.to_f
          s_t2 = mcTC[2 * k + 1]
          t2 = s_t2.to_f
          vtc = [t1, t2]
          @curObject.Appearance.TextureCoordinates << vtc
        end
      end
      if (bSaveThisElement)
        return super(name, prefix, uri)
      end
      return false
    end

    def on_characters(chars)
      debug = chars
      if ((@SurfaceBitMask & @POS) != 0)   #Es werden gerade Positionsangaben geparst
        @CharBuffer << chars
      end
      if ((@SurfaceBitMask & @POSLIST) != 0)   #Es werden gerade Positionsangaben als postList geparst
        @CharBuffer << chars
      elsif ((@SurfaceBitMask & @TEXTUREMAP) != 0)  #Es wird gerade der Texturnamen geparst
        @CharBuffer << chars
        super(chars)
      elsif ((@SurfaceBitMask & @TEXTURECOORDINATES) != 0)  #Es werden gerade Texturkoordinaten geparst
        @CharBuffer << chars
        super(chars)
      else    #die Charakters werden nicht geparst, nur gespeichert
        super(chars)
      end
    end
  end
end