require 'libxml'
require 'xml'
module FHGelsenkirchen
  class Validator
    class Handler
      include XML::SaxParser::Callbacks

      def initialize(sXmlFile, sketchup)
        @enable_sketchup = sketchup
        if(sketchup)
          @entities = Sketchup.active_model.entities
        end
        @nr_exterior = 0
        @nr_interior = 0
        @notplanar = 0
        @roofcount = 0
        @wallcount = 0
        @groundcount = 0
        # @log = File.new(File.dirname(sXmlFile) + "\\#{File.basename(sXmlFile, ".xml")}_ErrorLog.txt", "w")
        begin
          @log = File.new(File.dirname(sXmlFile) + "\\ErrorLog.txt", "w")
          dir = Dir.open(File.dirname(sXmlFile))
          dir.each { |filename|
            if(File.extname(filename) == ".xml")
              @surfacetype = ""
              @surfaceid = ""
              @poslistfailure = 0
              @posfailure = 0
              @ringnotclosed = 0
              @duplicatedpts = 0
              @lessthen3pts = 0
              @exterior = []
              @interiors = Hash.new
              @EMPTYSURFACE = 0
              @POS = 1
              @POSLIST = 2
              @TEXTUREMAP = 4
              @TEXTURECOORDINATES = 8
              @interior = false
              @polygon = false
              @SurfaceBitMask = @EMPTYSURFACE
              @CharBuffer = ""
              @linearringid = ""
              @polygonid = ""
              @notplanarids = []
              @duplicatedptsids = []
              @less3ptsids = []
              @notclosedids = []
              @eninteriorid = false
              @enexteriorid = false
              @interiorid = ""
              @buildingid = ""
              @polygonpos = 0
              @parser = XML::SaxParser.file(filename)
              @parser.callbacks= self
              @parser.parse()
              writeoutput(filename)
            end
           
          }
          @log << "Summary:\n"
          @log << "Number of Exteriors: #{@nr_exterior}\n"
          @log << "Number of Interiors: #{@nr_interior}\n"
          @log << "Not planar: #{@notplanar}\n"
          @log << "Not planar RoofSurfaces: #{@roofcount}\n"
          @log << "Not planar GroundSurfaces #{@groundcount}\n"
          @log << "Not planar WallSurfaces #{@wallcount}\n"
          @log.close
          #@parser = XML::SaxParser.file(sXmlFile)
          #@parser.callbacks= self
          #@parser.parse()
          #writeoutput(sXmlFile)
          #@log.close
          #name = File.dirname(sXmlFile) + "\\ErrorLog.txt"
          #Thread.new() { system("notepad \"#{name}\"") }
        rescue =>e
          puts e
          puts e.backtrace
        end
      end
      def writeoutput(sXmlFile)
        #@log << "Validator-Log #{File.basename(sXmlFile, ".xml")}\n"
        s = "#{sXmlFile}\n" if(@notplanarids.size > 0)
        #s = "Number of Exteriors: #{@nr_exterior}\n"
        #s << "Number of Interiors: #{@nr_interior}\n"
        #s << "Not planar: #{@notplanar}\n"
        #s << "Ring not Closed: #{@ringnotclosed}\n"
        #s << "Duplicated Points in Polygon: #{@duplicatedpts}\n"
        #s << "Polygons with less then 3 points: #{@lessthen3pts}\n"
        puts s if(@notplanarids.size > 0)
        @log << s if(@notplanarids.size > 0)
        #@log << "Not planar Polgon-IDs\n"
        @log << @notplanarids
        #@log << "Duplicated Points in Polygon-IDs\n"
        #@log << @duplicatedptsids
        #@log << "Polygon-IDs less then 3 points\n"
        #@log << @less3ptsids
        #@log << "Polygon not closed IDs\n"
        #@log << @notclosedids

      end
      def on_start_element_ns(name, attributes, prefix, uri, namespaces)
        case (name)
        when "LinearRing"
          begin
            if(@enexteriorid == true)
              @linearringid = attributes.fetch("id")
            else
              @interiorid = attributes.fetch("id")
              @interiors[@interiorid] = []
            end
          rescue => e
          end
        when "Polygon"
          @polygonpos += 1
          @polygon = true
          begin
            @polygonid = attributes.fetch("id")
          rescue => e
          end
        when "pos"
          @SurfaceBitMask |= @POS if(@polygon == true)
        when "posList"
          @SurfaceBitMask |= @POSLIST if(@polygon == true)
        when "interior"
          @eninteriorid = true
          @interior = true
          #@interiors[@interiorid] == []
        when "exterior"
          @enexteriorid = true
        when "Building"
          @polygonpos = 0
          begin
            @buildingid = attributes.fetch("id")
          rescue => e
          end
        when "GroundSurface"
		  @polygonpos = 0
          @surfacetype = "GroundSurface"
          begin
            @surfaceid = attributes.fetch("id")
          rescue => e
          end
        when "RoofSurface"
		  @polygonpos = 0
          @surfacetype = "RoofSurface"
          begin
            @surfaceid = attributes.fetch("id")
          rescue => e
          end
        when "WallSurface"
		  @polygonpos = 0
          @surfacetype = "WallSurface"
          begin
            @surfaceid = attributes.fetch("id")
          rescue => e
          end
        end
      end

      def on_end_element_ns(name, prefix, uri)
        case (name)
        when "GroundSurface"
          @surfaceid = ""
        when "RoofSurface"
          @surfaceid = ""
        when "WallSurface"
          @surfaceid = ""
        when "Building"
          @buildingid = ""
        when "Polygon"
          begin
            @nr_exterior += 1 if(@exterior.size > 1)
            @nr_interior += @interiors.size
            if(@exterior[0] != @exterior[-1])
              @ringnotclosed += 1
              if(@linearringid != "")
                @notclosedids << @linearringid  + "\n"
              elsif(@polygonid != "")
                @notclosedids << @polygonid  + "\n"
              end
            end
            @exterior.delete_at(@exterior.size - 1)
            @interiors.each_pair do |key,interior|
              @ringnotclosed += 1 if(interior[0] != interior[-1])
              interior.delete_at(interior.size - 1)
              if(interior.uniq! != nil)
                @duplicatedpts += 1
                @duplicatedptsids << key  + "\n" if(key != "")
              end
              if(interior.size < 3)
                @lessthen3pts += 1
                @less3ptsids << key  + "\n" if(key != "")
              end
            end
            if(@exterior.uniq! != nil)
              if(@linearringid != "")
                @duplicatedptsids << @linearringid  + "\n"
              elsif(@polygonid != "")
                @duplicatedptsids << @polygonid  + "\n"
              end
              @duplicatedpts += 1
            end
            if(@exterior.size < 3)
              if(@linearringid != "")
                @less3ptsids << @linearringid  + "\n"
              elsif(@polygonid != "")
                @less3ptsids << @polygonid  + "\n"
              end
              @lessthen3pts += 1
            end
            if(@enable_sketchup and @exterior.size > 2)
              begin
                face = @entities.add_face(@exterior)
                @entities.clear!
              rescue => e
                if(e.message == "Points are not planar")
                  @notplanar += 1
                  if(@polygonid != "")
                    @notplanarids << @polygonid  + "\n" if(@polygonid != "")
                  else
                    @notplanarids << "#{@buildingid} #{@surfaceid} Polygon #{@polygonpos} #{@surfacetype}\n"
                  end
                  case @surfacetype
                  when "GroundSurface"
                    @groundcount += 1
                  when "RoofSurface"
                    @roofcount += 1
                  when "WallSurface"
                    @wallcount += 1
                  end

                end
              end
            end
          rescue => err
            puts err.message
            puts err.backtrace
            return
          end
          @polygon = false
          @linearringid = ""
          @polygonid = ""
          @exterior = []
          @interiors = Hash.new
        when "pos"
          if(@polygon == true)
            begin
              pos = @CharBuffer.split(' ')
              if (pos.length % 3 != 0)
                @posfailure += 1
                puts "Parseerror: Count of \"Pos\" is not 3 (" + @CharBuffer +")"
              end
              if(@interior == true)
                @interiors[@interiorid] << [pos[0].to_f, pos[1].to_f, pos[2].to_f]
              else
                @exterior << [pos[0].to_f, pos[1].to_f, pos[2].to_f]
              end
              @SurfaceBitMask &= ~@POS
              @CharBuffer = ""
            rescue => err
              puts err
              puts err.backtrace
            end
          end
        when "posList"
          if(@polygon == true)
            mc_p_l = @CharBuffer.split(' ')
            if (mc_p_l.length%3 != 0)
              @poslistfailure += 1
              puts "posList failure: pos count is not a power of 3"
            end
            tmp = []
            for i in 0..(mc_p_l.length/3)-1
              tmp << [mc_p_l[3*i].to_f, mc_p_l[3*i + 1].to_f, mc_p_l[3*i + 2].to_f]
            end
            if(@interior == true)
              @interiors[@interiorid] = tmp
            else
              @exterior = tmp
            end
            @SurfaceBitMask &= ~@POSLIST
            @CharBuffer = ""
          end
        when "interior"
          @interior = false
          @eninteriorid = false
        when "exterior"
          @enexteriorid = false
        end
      end

      def on_characters(chars)
        if ((@SurfaceBitMask & @POS) != 0)   #Es werden gerade Positionsangaben geparst
          @CharBuffer << chars
        elsif ((@SurfaceBitMask & @POSLIST) != 0)   #Es werden gerade Positionsangaben als postList geparst
          @CharBuffer << chars
        end
      end
    end
    def validate(sXmlFile, sketchup)
      Handler.new(sXmlFile, sketchup)
    end
  end
end

#validator = FHGelsenkirchen::Validator.new
#validator.validate("E:\\CityGML Modelle\\Benfer Bottrop\\Bottrop.xml", false)